/////////////////////////////////////////////////////////////////////////////
//
// Windows MFC Glk Libraries
//
// GlkWindow
// Text buffer Glk windows
//
/////////////////////////////////////////////////////////////////////////////

#ifndef WINGLK_WINDOW_TEXTBUFFER_H_
#define WINGLK_WINDOW_TEXTBUFFER_H_

#include "GlkGraphic.h"
#include "GlkStyle.h"
#include "GlkWindow.h"

extern "C"
{
#include "glk.h"
}

/////////////////////////////////////////////////////////////////////////////
// Forward declarations
/////////////////////////////////////////////////////////////////////////////

class CWinGlkDC;

/////////////////////////////////////////////////////////////////////////////
// Text buffer Glk windows
/////////////////////////////////////////////////////////////////////////////

class CWinGlkWndTextBuffer : public CWinGlkWnd
{
	DECLARE_DYNAMIC(CWinGlkWndTextBuffer);

public:
	CWinGlkWndTextBuffer(glui32 Rock);
	virtual ~CWinGlkWndTextBuffer();

	virtual void SetActiveWindow(void);

	virtual void InitDC(CWinGlkDC& dc, CDC* pdcCompat = NULL);
	virtual bool CheckMorePending(void);
	virtual int GetFontHeight(void);

	virtual void SizeWindow(CRect* pSize);
	virtual void ClearWindow(void);
	virtual void PutCharacter(unsigned char c);

	virtual void GetSize(int& iWidth, int& iHeight);
	virtual void GetNeededSize(int iSize, int& iWidth, int& iHeight);

	virtual void StartLineEvent(char* pBuffer, int iMaxLength, int iStartLength);
	virtual void StartLinkEvent(void);
	virtual void EndLinkEvent(void);
	virtual void InputChar(UINT uiChar);

	virtual bool MouseClick(CPoint& Click);
	virtual unsigned int GetLinkAtPoint(const CPoint& Point);

	virtual void SetStyle(int iStyle);
	virtual int GetStyle(void);
	virtual CWinGlkStyle* GetStyle(int iStyle);
	virtual void SetHyperlink(unsigned int iLink);
	virtual bool DistinguishStyles(int iStyle1, int iStyle2);
	virtual bool MeasureStyle(int iStyle, int iHint, glui32* pResult);

	virtual bool DrawGraphic(CWinGlkGraphic* pGraphic,
		int iValue1, int iValue2, int iWidth, int iHeight, bool& bDelete);
	void InsertFlowBreak(void);

	virtual void Scrollback(void);

protected:
	//{{AFX_MSG(CWinGlkWndTextBuffer)
	afx_msg void OnPaint();
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()

protected:

	class CHyperlink : public CRect
	{
	public:
		CHyperlink(const CRect& Rect, unsigned int iLink);
		CHyperlink();

		unsigned int m_iLink;
	};

	class CPaintInfo
	{
	public:
		CPaintInfo(int iLeft, int iTop, int iWidth, int iHeight, CWinGlkDC& DeviceContext, CArray<CHyperlink,CHyperlink&>& Hyperlinks);
		~CPaintInfo();

		void MarginAdd(CWinGlkGraphic* pGraphic);
		void MarginPaint(CWinGlkGraphic* pGraphic, int iIndent);
		void MarginPaintPrev(CWinGlkGraphic* pGraphic, int& iOffset, int iIndent);
		void NextLine(int iStep, bool bDown);
		bool HasMarginGraphic(void);
		int GetMaxMarginHeight(void);
		void CheckHyperlink(const CRect& Rect);
		void DrawGraphic(CWinGlkGraphic* pGraphic, int iLeft, int iTop);

	public:
		int m_iLeft;
		int m_iTop;
		int m_iWidth;
		int m_iHeight;
		CWinGlkDC& m_DeviceContext;

	protected:
		class CMarginInsert
		{
		public:
			CMarginInsert();

		public:
			int m_iHeightLeft;
			int m_iGraphicWidth;
			bool m_bOnLeft;
		};

		CArray<CMarginInsert,CMarginInsert&> m_Margins;

	protected:
		CArray<CHyperlink,CHyperlink&>& m_Hyperlinks;
	};

	class CLineFormat
	{
	public:
		CLineFormat();
		~CLineFormat();

	public:
		int m_iFirstCharacter;
		int m_iLastCharacter;
		int m_iMaxAboveBaseline;
		int m_iMaxBelowBaseline;
		int m_iLineLength;
	};

	class CParagraph
	{
	public:
		CParagraph(int iStyle, unsigned int iLink);
		~CParagraph();

		void AddCharacter(unsigned char c);
		bool AddGraphic(CWinGlkGraphic* pGraphic);
		int GetLength();
		void SetInitialStyle(int iStyle);
		void SetInitialLink(unsigned int iLink);
		void AddStyleChange(int iStyle);
		void AddLinkChange(unsigned int iLink);

		void ClearFormatting(void);
		void Format(CPaintInfo& Info);
		void Update(CPaintInfo& Info, int& iOffset, CWinGlkWnd* pWnd);
		bool Paint(CPaintInfo& Info, int& iFinalLeft, int& iFinalTop, bool bMark);
		void SetAsShown(void);

		int GetCharCount(void) const;
		int CopyChars(LPTSTR pszBuffer) const;
		int GetHeight(void) const;
		int GetLastShown(void) const;
		int LastShownHeight(void) const;
		bool ContainsText(void) const;

		void AddInteger(unsigned int iValue);
		unsigned int GetInteger(int iPos) const;

		void Speak(void);

	protected:
		enum TextCodes
		{
			StyleChange = 1,
			LinkChange,
			InlineGraphic,
			FlowBreak,
		};

	protected:
		bool TestLineLength(CPaintInfo& Info, CString& strLine,
			CSize& Size, int& iLeftEdge, int iIndent, bool bFinal,
			int& iIndex, int& iLastBreak, int& iLastPossible,
			int& iLastLength, int& iMaxUp, int& iMaxDown, bool& bExit);
		void TextOut(CLineFormat* pFormat, CPaintInfo& Info,
			int& iLeft, char* pBuffer, int& iBufferPos);
		void SetLastShown(CPaintInfo& Info, int iIndex);
		int JustifyLength(CLineFormat* pFormat, CPaintInfo& Info,
			char* pBuffer, int& iSpaces);

	protected:
		CArray<unsigned char,unsigned char> m_Text;
		CArray<CLineFormat*,CLineFormat*> m_Formatting;
		CArray<CWinGlkGraphic*,CWinGlkGraphic*> m_InlineGraphics;
		CArray<CWinGlkGraphic*,CWinGlkGraphic*> m_MarginGraphics;
		int m_iInitialStyle;
		unsigned int m_iInitialLink;
		int m_iLastShown;
		bool m_bSpoken;
		static const int m_iIndentStep;
	};

protected:
	void Paint(bool bMark);
	void CheckDeleteOldText(int iTop, int iClientHeight);
	void PreparePaintInfo(CPaintInfo& Info, CRect& ClientArea, int& iParagraph, int& iOffset);
	void AddNewParagraph(void);

	CArray<CParagraph*,CParagraph*> m_TextBuffer;
	CArray<CHyperlink,CHyperlink&> m_Hyperlinks;
	CArray<unsigned char,unsigned char> m_ScrollBuffer;
	CWinGlkStyles m_Styles;
	int m_iCurrentStyle;
	unsigned int m_iCurrentLink;
	bool m_bCheckDeleteText;
	bool m_bMorePending;

public:
	static void SetStyleHint(int iStyle, int iHint, glsi32 Value);
	static void ClearStyleHint(int iStyle, int iHint);
	static CWinGlkStyles* GetDefaultStyles(void) { return &m_DefaultTextBufferStyles; }

protected:
	static CWinGlkStyles m_DefaultTextBufferStyles;
	static CString m_strMore;
};

/////////////////////////////////////////////////////////////////////////////
// Device context class
/////////////////////////////////////////////////////////////////////////////

class CWinGlkBufferDC : public CWinGlkDC
{
public:
	CWinGlkBufferDC(CWinGlkWnd* pWnd) : CWinGlkDC(pWnd) {};
	virtual ~CWinGlkBufferDC() {};

	virtual void GetFonts(LOGFONT*& pTextFont, LOGFONT*& pSizeFont);
	virtual void SetFontStyles(LOGFONT& Font);
	virtual int GetStyleFontSize(void);
};

#endif // WINGLK_WINDOW_TEXTBUFFER_H_
